

PUBLIC vortex_256_asm
PUBLIC vortex_512_asm

_TEXT	SEGMENT

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
; Our prototype
; void vortex_256_asm(unsigned char *hash, unsigned char *data, unsigned char *rcon, int length)	
;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;


vortex_256_asm PROC

	;rcx holds the pointer to the hash 
	;rdx holds the pointer to the data
	;r8  holds the pointer to the rcon constants
	;r9  holds the pointer to the length in blocks
	
	
	;first we load the hash in the register pair <xmm1:xmm0>
	
	movdqu		xmm0,	XMMWORD PTR [rcx]
	movdqu		xmm1,	XMMWORD PTR [rcx+16]
	
	;then we load the rcon constants in the registers xmm13, xmm14, xmm15
	
	movdqu		xmm13,	XMMWORD PTR [r8]
	movdqu		xmm14,	XMMWORD PTR [r8+16]
	movdqu		xmm15,	XMMWORD PTR [r8+32]
	
	;xmm12 holds the constant zero
	pxor		xmm12,	xmm12

	
vortex_block_loop:
	mov			r10,	4
vortex_word_loop:
	;first we load the data	into the register xmm11
	movdqu		xmm11,	XMMWORD PTR [rdx]

	;1st block: move the data to xmm4, xmm5
	movdqu		xmm4,	xmm11
	movdqu		xmm5,	xmm11
	
	;do the modified key schedule
	paddq		xmm0,	xmm13
	paddq		xmm1,	xmm13
	
	;permutation + sbox can be implemented using the future pshufb + aeseclast instructions
	;we simulate these using pxor, pmuludq
	
	;pshufb		xmm0
	pxor		xmm0,	xmm13	;1 clock stand-in
	;pshufb		xmm1
	pxor		xmm1,	xmm13
	;aesenclast	xmm0,	xmm12
	pmuludq		xmm0,	xmm12	;3 clock latency, 1 clock throughput stand it
	;aesenclast	xmm1,	xmm12
	pmuludq		xmm1,	xmm12
	
	;the keys for the first round are in the registers xmm1:xmm0
	;to pipeline the execution of the aes round instructions we begin preparing
	;the key schedule for the next round
	
	movdqu		xmm2,	xmm0
	movdqu		xmm3,	xmm1
	paddq		xmm2,	xmm14
	paddq		xmm3,	xmm14
	;pshufb		xmm2
	pxor		xmm2,	xmm14	;1 clock stand-in
	;pshufb		xmm3
	pxor		xmm3,	xmm14
	
	;now we issue four aes instructions 2 for the first round and 2 for the next key schedule
	;aesenc		xmm4,	xmm0
	pmuludq		xmm4,	xmm0
	;aesenc		xmm5,	xmm1
	pmuludq		xmm5,	xmm1	
	;aesenclast	xmm2,	xmm12
	pmuludq		xmm2,	xmm12	;3 clock latency, 1 clock throughput stand it
	;aesenclast	xmm3,	xmm12
	pmuludq		xmm3,	xmm12
	
	;first round done - key schedule for second round prepared
	
	;we begin preparing the key schedule for the third round
	
	movdqu		xmm0,	xmm2
	movdqu		xmm1,	xmm3
	paddq		xmm0,	xmm15
	paddq		xmm1,	xmm15
	;pshufb		xmm0
	pxor		xmm0,	xmm15	;1 clock stand-in
	;pshufb		xmm1
	pxor		xmm1,	xmm15
	
	;now we issue four aes instructions 2 for the first round and 2 for the next key schedule
	;aesenc		xmm4,	xmm2
	pmuludq		xmm4,	xmm2
	;aesenc		xmm5,	xmm3
	pmuludq		xmm5,	xmm3	
	;aesenclast	xmm0,	xmm12
	pmuludq		xmm0,	xmm12	;3 clock latency, 1 clock throughput stand it
	;aesenclast	xmm1,	xmm12
	pmuludq		xmm1,	xmm12
	
	;second round done - key schedule for third round prepared
	
	;aesenc		xmm4,	xmm0
	pmuludq		xmm4,	xmm0
	;aesenc		xmm5,	xmm1
	pmuludq		xmm5,	xmm1
	
	;last round done!	
	pxor		xmm4,	xmm11	;matyas-mayer-oseyas
	pxor		xmm5,	xmm11		
	
	;now we start the merging
	
	movdqu		xmm1, xmm4
	;pclmulqdq	xmm1, xmm5, 0x10 ;xmm1 holds the outer product
	pmuludq		xmm1, xmm5
	movdqu		xmm0, xmm4
	;pclmulqdq	xmm0, xmm5, 0x01 ;xmm0 holds the inner product
	pmuludq		xmm0, xmm5
	
	movdqu		xmm2,	xmm0
	movdqu		xmm3,	xmm1
	shufpd		xmm0,	xmm3, 0
	shufpd		xmm1,	xmm2, 0
	pxor		xmm0,	xmm4
	paddq		xmm1,	xmm5
	;we are done!
	
	add			rdx,	16
	dec			r10
	jnz			vortex_word_loop
	dec			r9
	jnz			vortex_block_loop
	
	;we load the hash back
	movdqu		XMMWORD PTR [rcx], xmm0
	movdqu		XMMWORD PTR [rcx+16], xmm1

	RET
		
vortex_256_asm ENDP	


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
; Our prototype
; void vortex_512_asm(unsigned char *hash, unsigned char *data, unsigned char *rcon, int length)	
;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;


vortex_512_asm PROC

	;rcx holds the pointer to the hash 
	;rdx holds the pointer to the data
	;r8  holds the pointer to the rcon constants
	;r9  holds the pointer to the length in blocks
	
	
	;first we load the hash in the register set <xmm3:xmm2:xmm1:xmm0>
	
	movdqu		xmm0,	XMMWORD PTR [rcx]
	movdqu		xmm1,	XMMWORD PTR [rcx+16]
	movdqu		xmm2,	XMMWORD PTR [rcx+32]
	movdqu		xmm3,	XMMWORD PTR [rcx+48]
	
	;xmm15 holds the constant zero
	pxor		xmm15,	xmm15

	
vortex_512_block_loop:
	mov			r10,	4
	
vortex_512_word_loop:

	;first we load the data	into the register pair <xmm14:xmm13>
	movdqu		xmm13,	XMMWORD PTR [rdx]
	movdqu		xmm14,	XMMWORD PTR [rdx+16]
	

	;1st block: move the data to xmm8, xmm9, xmm10, xmm11
	movdqu		xmm8,	xmm13
	movdqu		xmm9,	xmm14
	movdqu		xmm10,	xmm13
	movdqu		xmm11,	xmm14
	
	
	;do the modified key schedule
	paddq		xmm0,	XMMWORD PTR [r8]
	paddq		xmm1,	XMMWORD PTR [r8+16]
	paddq		xmm2,	XMMWORD PTR [r8]
	paddq		xmm3,	XMMWORD PTR [r8+16]
	
	
	;permutation + sbox can be implemented using the future pshufb + aeseclast instructions
	;we simulate these using pxor, pmuludq
	
	;pshufb		xmm0
	pxor		xmm0,	xmm13	;1 clock stand-in
	;pshufb		xmm1
	pxor		xmm1,	xmm13
	;pshufb		xmm2
	pxor		xmm2,	xmm13
	;pshufb		xmm3
	pxor		xmm3,	xmm13
	
	;aesenclast	xmm0,	xmm15
	pmuludq		xmm0,	xmm15	;3 clock latency, 1 clock throughput stand it
	;aesenclast	xmm1,	xmm15
	pmuludq		xmm1,	xmm15
	;aesenclast	xmm2,	xmm15
	pmuludq		xmm2,	xmm15
	;aesenclast	xmm3,	xmm15
	pmuludq		xmm3,	xmm15
	
	;the keys for the first round are in the registers xmm3:xmm2 and xmm1:xmm0
	;to pipeline the execution of the aes round instructions we begin preparing
	;the key schedule for the next round
	
	movdqu		xmm4,	xmm0
	movdqu		xmm5,	xmm1
	movdqu		xmm6,	xmm2
	movdqu		xmm7,	xmm3
	
	paddq		xmm4,	XMMWORD PTR [r8+32]
	paddq		xmm5,	XMMWORD PTR [r8+48]
	paddq		xmm6,	XMMWORD PTR [r8+32]
	paddq		xmm7,	XMMWORD PTR [r8+48]
	
	
	;pshufb		xmm4
	pxor		xmm4,	xmm13	;1 clock stand-in
	;pshufb		xmm5
	pxor		xmm5,	xmm13
	;pshufb		xmm6
	pxor		xmm6,	xmm13
	;pshufb		xmm7
	pxor		xmm7,	xmm13
	
	;for rijndael 256 we need to do 2 256-wide shuffle operations - we simulate them 
	;with por
	por			xmm8,	xmm10
	por			xmm9,	xmm11
	
	;now we issue 8 aes instructions 4 for the first round and 4 for the next key schedule
	;aesenc		xmm8,	xmm0
	pmuludq		xmm8,	xmm0
	;aesenc		xmm9,	xmm1
	pmuludq		xmm9,	xmm1
	;aesenc		xmm10,	xmm2
	pmuludq		xmm10,	xmm2
	;aesenc		xmm11,	xmm3
	pmuludq		xmm11,	xmm3
	
	;aesenclast	xmm4,	xmm15
	pmuludq		xmm4,	xmm15	;3 clock latency, 1 clock throughput stand it
	;aesenclast	xmm5,	xmm15
	pmuludq		xmm5,	xmm15
	;aesenclast	xmm6,	xmm15
	pmuludq		xmm6,	xmm15
	;aesenclast	xmm7,	xmm15
	pmuludq		xmm7,	xmm15
	
	;first round done - key schedule for second round prepared
	
	;we begin preparing the key schedule for the third round
	
	movdqu		xmm0,	xmm4
	movdqu		xmm1,	xmm5
	movdqu		xmm2,	xmm6
	movdqu		xmm3,	xmm7
	
	paddq		xmm0,	XMMWORD PTR [r8+64]
	paddq		xmm1,	XMMWORD PTR [r8+80]
	paddq		xmm2,	XMMWORD PTR [r8+64]
	paddq		xmm3,	XMMWORD PTR [r8+80]
	
	;pshufb		xmm0
	pxor		xmm0,	xmm13	;1 clock stand-in
	;pshufb		xmm1
	pxor		xmm1,	xmm13
	;pshufb		xmm2
	pxor		xmm2,	xmm13
	;pshufb		xmm3
	pxor		xmm3,	xmm13
	
	;for rijndael 256
	pand		xmm8,	xmm10
	pand		xmm9,	xmm11
	
	;now we issue 8 aes instructions 4 for the first round and 4 for the next key schedule
	
	;aesenc		xmm8,	xmm4
	pmuludq		xmm8,	xmm4
	;aesenc		xmm9,	xmm5
	pmuludq		xmm9,	xmm5
	;aesenc		xmm10,	xmm6
	pmuludq		xmm10,	xmm6
	;aesenc		xmm11,	xmm7
	pmuludq		xmm11,	xmm7
	
	;aesenclast	xmm0,	xmm15
	pmuludq		xmm0,	xmm15	;3 clock latency, 1 clock throughput stand it
	;aesenclast	xmm1,	xmm15
	pmuludq		xmm1,	xmm15
	;aesenclast	xmm2,	xmm15
	pmuludq		xmm2,	xmm15
	;aesenclast	xmm3,	xmm15
	pmuludq		xmm3,	xmm15
	
	;second round done - key schedule for third round prepared
	
	;rijndael  256
	por			xmm8,	xmm10
	por			xmm9,	xmm11
	
	;aesenc		xmm8,	xmm0
	pmuludq		xmm8,	xmm0
	;aesenc		xmm9,	xmm1
	pmuludq		xmm9,	xmm1
	;aesenc		xmm10,	xmm2
	pmuludq		xmm10,	xmm2
	;aesenc		xmm11,	xmm3
	pmuludq		xmm11,	xmm3
	
	;last round done!	
	pxor		xmm8,	xmm13	;matyas-mayer-oseyas
	pxor		xmm9,	xmm14
	pxor		xmm10,	xmm13
	pxor		xmm11,	xmm14		
	
	;now we start the merging
	
	;we multiply carry-less xmm9 with xmm10, place the result in <xmm3:xmm0>
	
	movdqu		xmm0, xmm9
	;pclmulqdq	xmm0, xmm10, 0x00 ;xmm0 holds a0*b0
	pmuludq		xmm0, xmm10
	movdqu		xmm4, xmm9
	;pclmulqdq	xmm4, xmm10, 0x10 ;xmm4 holds a0*b1
	pmuludq		xmm4, xmm10
	movdqu		xmm5, xmm9
	;pclmulqdq	xmm5, xmm10, 0x01 ;xmm5 holds a1*b0
	pmuludq		xmm5, xmm10
	movdqu		xmm3, xmm9
	;pclmulqdq	xmm3, xmm10, 0x11 ;xmm3 holds a1*b1
	pmuludq		xmm3, xmm10
	pxor		xmm4, xmm5		 ;xmm4 holds a0*b1 + a1*b0
	movdqu		xmm5, xmm4		 ; move the content of xmm4 to xmm5
	psrldq		xmm4, 8			 ; shift xmm4 by 64 bits to the right
	pslldq		xmm5, 8			 ; shift xmm5 by 64 bits to the left	
	pxor		xmm0, xmm5
	pxor		xmm3, xmm4	     ; register pair <xmm3:xmm0> holds the result of 
							     ; the carry-less multiplication of xmm9 by xmm10
	
	;we also multiply carry-less xmm8 with xmm11, place the result in <xmm1:xmm2>
	
	movdqu		xmm2, xmm8
	;pclmulqdq	xmm2, xmm11, 0x00 ;xmm2 holds a0*b0
	pmuludq		xmm2, xmm11
	movdqu		xmm4, xmm8
	;pclmulqdq	xmm4, xmm11, 0x10 ;xmm4 holds a0*b1
	pmuludq		xmm4, xmm11
	movdqu		xmm5, xmm8
	;pclmulqdq	xmm5, xmm11, 0x01 ;xmm5 holds a1*b0
	pmuludq		xmm5, xmm11
	movdqu		xmm1, xmm8
	;pclmulqdq	xmm1, xmm11, 0x11 ;xmm1 holds a1*b1
	pmuludq		xmm1, xmm11
	pxor		xmm4, xmm5		 ;xmm4 holds a0*b1 + a1*b0
	movdqu		xmm5, xmm4		 ; move the content of xmm4 to xmm5
	psrldq		xmm4, 8			 ; shift xmm4 by 64 bits to the right
	pslldq		xmm5, 8			 ; shift xmm5 by 64 bits to the left	
	pxor		xmm2, xmm5
	pxor		xmm1, xmm4	     ; register pair <xmm1:xmm2> holds the result of 
							     ; the carry-less multiplication of xmm8 by xmm11
		
	pxor		xmm0,	xmm8
	pxor		xmm1,	xmm9	
	paddq		xmm2,	xmm10
	paddq		xmm3,	xmm11
	;we are done!
	
	add			rdx,	32
	dec			r10
	jnz			vortex_512_word_loop
	dec			r9
	jnz			vortex_512_block_loop
	
	;we load the hash back
	movdqu		XMMWORD PTR [rcx], xmm0
	movdqu		XMMWORD PTR [rcx+16], xmm1

	RET
		
vortex_512_asm ENDP	


_TEXT ENDS
END
